import unittest
from pyshamir import split, combine
from base64 import b64encode, b64decode

class TestSplit(unittest.TestCase):
  secret = b64decode('a+m4G0kEkKDQK4MFGz7L0vLz5oViQkDSLThiC4zDRZU=')
  def test_split(self):
    try:
      parts = split(self.secret, 5, 3)
      self.assertEqual(len(parts),5)
    except ValueError as ve:
      self.fail(ve.args[0])
    for part in parts:
      self.assertEqual(len(part), len(self.secret)+1)

    # test that the parts are different
    for i in range(len(parts) - 1):
      self.assertNotEqual(parts[i].hex(), parts[i + 1].hex())

  def test_split_invalid(self):
    with self.assertRaisesRegex(ValueError, 'Parts and threshold must be greater than 1'):
      split(self.secret, 0, 0)

    with self.assertRaisesRegex(ValueError, 'Parts must be greater than threshold'):
      split(self.secret, 2, 3)

    with self.assertRaisesRegex(ValueError, 'Parts must be less than 256'):
      split(self.secret, 1000, 3)

    with self.assertRaisesRegex(ValueError, 'Secret must be at least 1 byte long'):
      split(None, 5, 3)

    with self.assertRaisesRegex(ValueError, 'Secret must be at least 1 byte long'):
      split(bytearray(b''), 5, 3)

class TestCombine(unittest.TestCase):
  secret = b64decode('esfX3MUC++BrcwkiwsKtK6M5Pi5yvuc/A/6LweWJ5FA=')

  def test_combine(self):
    try:
      parts = split(self.secret, 5, 3)
    except ValueError as ve:
      self.fail(ve.args[0])

    # brute force test if all the possible combinations of keys
    # result in the same secret
    for i in range(5):
      for j in range(5):
        if j == i:
          continue
        for k in range(5):
          if k == i or k == j:
            continue
          int_parts = [parts[i], parts[j], parts[k]]
          try:
            recomb = combine(int_parts)
            self.assertEqual(b64encode(recomb), b64encode(self.secret))
          except ValueError as ve:
            self.fail(ve.args[0])

  def test_combine_invalid(self):
    with self.assertRaisesRegex(ValueError, 'Not enough parts to combine'):
      combine(bytearray())
    with self.assertRaisesRegex(ValueError, 'Not enough parts to combine'):
      combine(None)

    #check part length mismatch
    parts = [bytearray(b'ab'),bytearray(b'abc')]
    with self.assertRaisesRegex(ValueError, 'Parts are not the same length'):
      combine(parts)

    #check part length too small
    parts = [bytearray(b'a'),bytearray(b'b')]
    with self.assertRaisesRegex(ValueError, 'Part is too short'):
      combine(parts)

    #duplicate
    parts = [bytearray(b'ab'),bytearray(b'ab')]
    with self.assertRaisesRegex(ValueError, 'Duplicate sample'):
      combine(parts)

if __name__ == '__main__':
    unittest.main()



import secrets
from ctypes import c_uint8

def add(a, b)->int:
    """
    Adds two numbers in the finite field GF(256)
    """
    out = a ^ b
    return out


def inverse(a):
    b = mul(a, a)
    c = mul(a, b)
    b = mul(c, c)
    b = mul(b, b)
    c = mul(b, c)
    b = mul(b, b)
    b = mul(b, b)
    b = mul(b, c)
    b = mul(b, b)
    b = mul(a, b)
    return mul(b, b)


def mul(a, b):
    a = c_uint8(a)
    b = c_uint8(b)
    r = c_uint8(0)
    i = 8

    while i > 0:
        i -= 1
        p1 = (-(c_uint8(b.value >> i).value & 1) & a.value)
        p2 = (-(c_uint8(r.value >> 7).value) & 0x1B)
        p3 = c_uint8(2* r.value)
        r = c_uint8(p1 ^ p2 ^ p3.value)

    return r.value


def div(a, b)->int:
    a = c_uint8(a)
    b = c_uint8(b)

    if b.value == 0:
        raise ZeroDivisionError("Divide by zero")
    ret = mul(a.value,inverse(b.value))

    if a.value == 0:
        return 0

    return ret

class Polynomial:
    """
    Takes N sample points and returns the value of the polynomial at x using Lagrange interpolation
    """
    def __init__(self, degree):
        self.coefficients = bytearray(degree + 1)

    def evaluate(self, x)->int:
        # origin case
        if x == 0:
            return self.coefficients[0]

        # compute the polynomial value using Horner's method
        degree = len(self.coefficients) - 1
        out = self.coefficients[degree]
        for i in range(degree - 1, -1, -1):
            coeff = self.coefficients[i]
            out = add(mul(out, x), coeff)
        return out

def make_polynomial(intercept, degree)->Polynomial:
    """
    Creates a random polynomial with the given intercept and degree
    :param intercept:
    :param degree:
    :return:
    """
    polynomial_instance = Polynomial(degree)

    # Set the intercept
    polynomial_instance.coefficients[0] = intercept

    # assign random co-efficients to the polynomial
    polynomial_instance.coefficients[1:] = secrets.token_bytes(degree)

    return polynomial_instance

def interpolate_polynomial(x_samples, y_samples, x)->int:
    """
    Takes N sample points and returns the value of the polynomial at x using Lagrange interpolation
    """
    limit = len(x_samples)
    result = 0
    for i in range(limit):
        basis = 1
        for j in range(limit):
            if i != j:
                num = add(x, x_samples[j])
                den = add(x_samples[i], x_samples[j])
                term = div(num, den)
                basis = mul(basis, term)
        group = mul(y_samples[i], basis)
        result = add(result, group)
    return result

def generate_x_coordinates(n):
    x_coordinates = list(range(n))
    secrets.SystemRandom().shuffle(x_coordinates)
    return x_coordinates



import secrets
from ._utils import make_polynomial, interpolate_polynomial, generate_x_coordinates

def combine(parts: list) -> bytearray:
    """
    Takes a list of parts and returns the secret
    :param parts:
    :return:
    """
    # Verify enough parts are present
    if parts is None:
        raise ValueError("Not enough parts to combine")
    if len(parts) < 2:
        raise ValueError("Not enough parts to combine")

    # Verify all parts are all the same length
    first_part_len = len(parts[0])

    if first_part_len < 2:
        raise ValueError("Part is too short")

    for part in parts:
        if len(part) != first_part_len:
            raise ValueError("Parts are not the same length")

    # Create a buffer to store the reconstructed secret
    secret = bytearray(first_part_len - 1)

    # Buffer to store the samples
    x_samples = bytearray(len(parts))
    y_samples = bytearray(len(parts))

    # set the x value for each sample and ensure the no_sample values are the same , otherwise div() can be unhappy
    check_map = {}

    for i, part in enumerate(parts):
        samp = part[first_part_len - 1]
        if samp in check_map:
            raise ValueError("Duplicate sample")
        check_map[samp] = True
        x_samples[i] = samp

    # Reconstruct each  byte
    for idx, _ in enumerate(secret):
        for i, part in enumerate(parts):
            y_samples[i] = part[idx]

        # interpolate the polynomial and compute the vault at 0
        val = interpolate_polynomial(x_samples, y_samples, 0)

        # Evaluate the 0th value to get the intercept
        secret[idx] = val

    return secret

def split(secret: bytes, parts: int, threshold: int) -> list:
    """
    Takes a secret and splits it into parts
    :param secret:
    :param parts:
    :param threshold:
    :return:
    """
    # Sanity check the input
    if parts < 2 or threshold < 2:
        raise ValueError("Parts and threshold must be greater than 1")
    if parts < threshold:
        raise ValueError("Parts must be greater than threshold")
    if parts > 255:
        raise ValueError("Parts must be less than 256")
    if secret is None:
        raise ValueError("Secret must be at least 1 byte long")
    if len(secret) < 1:
        raise ValueError("Secret must be at least 1 byte long")

    # Generate random list of x coordinates
    x_coordinates = generate_x_coordinates(255)

    # Allocate the output array , initalize the final byte of the output with the offset.
    # This is used to ensure that the same secret can be split into different parts
    output = [bytearray() for _ in range(parts)]
    for i in range(len(output)):
        output[i] = bytearray(len(secret) + 1)
        output[i][len(secret)] = int(x_coordinates[i]) + 1

    for i, val in enumerate(secret):
        polynomial_instance = make_polynomial(val, int(threshold - 1))

        for j in range(parts):
            x = int(x_coordinates[j]) + 1
            y = polynomial_instance.evaluate(x)
            output[j][i] = y

    return output
